#include "WaveExtractor.hpp"
#include "Logger.hpp"
#include "WaveFileWriter.hpp"
#include <tchar.h>
#include <process.h>

WaveExtractor::WaveExtractor(const Logger* logger)
: file_(NULL), logger_(logger)
{
}

WaveExtractor::~WaveExtractor()
{
	close();
}

bool WaveExtractor::open(const TCHAR* filepath)
{
	PAVIFILE file;
	HRESULT result = AVIFileOpen(&file, filepath, OF_SHARE_DENY_WRITE, NULL);
	switch (result)
	{
	case AVIERR_BADFORMAT:
		logger_->error(_T("̓t@CJ܂Bt@CjĂ邩AtH[}bgsłB\n"));
		return false;

	case AVIERR_MEMORY:
		logger_->error(_T("̓t@CJ܂BȂȂ܂B\n"));
		return false;

	case AVIERR_FILEREAD:
		logger_->error(_T("̓t@CJ܂BǂݍݒɃfBXNG[܂B\n"));
		return false;

	case AVIERR_FILEOPEN:
		logger_->error(_T("̓t@CJ܂Bt@CJۂɃfBXNG[܂B\n"));
		return false;

	case REGDB_E_CLASSNOTREG:
		logger_->error(_T("̓t@CJ܂Bw肳ꂽt@C^CvnhAWXgɌ܂łB\n"));
		return false;

	default:
		if (FAILED(result))
		{
			logger_->error(_T("̓t@CJ܂BsȃG[܂B\n"));
			return false;
		}
	}

	file_ = file;

	logger_->info(_T("̓t@CJ܂B\n"));

	return true;
}

void WaveExtractor::close()
{
	if (file_ != NULL)
	{
		AVIFileRelease(file_);
		file_ = NULL;
		logger_->info(_T("̓t@C܂B\n"));
	}
}

bool WaveExtractor::extract(int streamNumber, const TCHAR* outputFilePath)
{
	// w̃I[fBIXg[J
	PAVISTREAM stream;
	HRESULT result = AVIFileGetStream(file_, &stream, streamtypeAUDIO, streamNumber);
	if (result == AVIERR_NODATA)
	{
		logger_->error(_T("I[fBIXg[J܂Bw肳ꂽXg[݂͑܂B\n"));
		return false;
	}
	else if (result == AVIERR_MEMORY)
	{
		logger_->error(_T("I[fBIXg[J܂BȂȂ܂B\n"));
		return false;
	}
	else if (FAILED(result))
	{
		logger_->error(_T("I[fBIXg[J܂BsȃG[܂B\n"));
		return false;
	}

	logger_->info(_T("I[fBIXg[J܂B\n"));

	// o{
	bool success = extractImpl(stream, outputFilePath);

	// I[fBIXg[
	AVIStreamRelease(stream);

	logger_->info(_T("I[fBIXg[܂B\n"));

	return success;
}

unsigned int WaveExtractor::waveWriteThread(void* param)
{
	SharedData* data = reinterpret_cast<SharedData*>(param);

	// ǂݍ݃XbhŃG[oȂ胋[v
	while (!data->error)
	{
		// ̎_̓ǂݍ݃J[\擾
		size_t readCursor = data->readCursor;

		// of[^邩
		if (readCursor > data->writer->getWritten())
		{
			// obt@̓ǂݎ\͈͂擾
			size_t start = data->writer->getWritten() % data->bufferLength;
			size_t end = readCursor % data->bufferLength;
			if (end <= start)
				end = data->bufferLength;

			// o
			if (!data->writer->write(&data->buffer[start], end - start))
			{
				data->error = true;
				break;
			}

			// oI𔻒
			if (data->writer->isComplete())
				break;

			// ǂݍޗ]nôœǂݍ݃XbhN
			SetEvent(data->readEvent);
		}
		else
		{
			// of[^̂œǂݍ݃XbhNđ҂
			SetEvent(data->readEvent);
			WaitForSingleObject(data->writeEvent, INFINITE);
		}
	}

	// ǂݍ݃XbhNďI
	SetEvent(data->readEvent);

	return 0;
}

bool WaveExtractor::extractImpl(PAVISTREAM stream, const TCHAR* outputFilePath)
{
	// I[fBIXg[̏擾
	AVISTREAMINFO  streamInfo;
	HRESULT result = AVIStreamInfo(stream, &streamInfo, sizeof(AVISTREAMINFO));
	if (FAILED(result))
	{
		logger_->error(_T("I[fBIXg[̎擾Ɏs܂B\n"));
		return false;
	}

	logger_->info(_T("I[fBIXg[擾܂B\n"));
	logger_->verbos(_T("TṽTCY: %d Byte\n"), streamInfo.dwSampleSize);
	logger_->verbos(_T("Tv: %d\n"), streamInfo.dwLength);


	// I[fBIXg[WavetH[}bg擾
	LONG formatSize = 0;
	result = AVIStreamReadFormat(stream, 0, NULL, &formatSize);
	if (FAILED(result))
	{
		logger_->error(_T("I[fBIXg[WavetH[}bg擾Ɏs܂B\n"));
		return false;
	}
	if (formatSize > sizeof(WAVEFORMATEX))
	{
		logger_->error(_T("I[fBIXg[WavetH[}bg擾Ɏs܂B\n"));
		return false;
	}

	WAVEFORMATEX format;
	ZeroMemory(&format, sizeof(WAVEFORMATEX));
	result = AVIStreamReadFormat(stream, 0, &format, &formatSize);
	if (FAILED(result))
	{
		logger_->error(_T("I[fBIXg[WavetH[}bg擾Ɏs܂B\n"));
		return false;
	}

	logger_->info(_T("I[fBIXg[WavetH[}bg擾܂B\n"));
	logger_->verbos(_T("`l: %d\n"), format.nChannels);
	logger_->verbos(_T("rbg: %d\n"), format.wBitsPerSample);
	logger_->verbos(_T("TvO[g: %d\n"), format.nSamplesPerSec);


	// g`f[^ǂݍWavet@CƂĕۑ
	bool complete = false; // tO

	SharedData* data = new SharedData(format, streamInfo.dwSampleSize, streamInfo.dwSampleSize * streamInfo.dwLength, logger_);
	if (data->initialize(outputFilePath))
	{
		HANDLE thread = reinterpret_cast<HANDLE>(_beginthreadex(NULL, 0, waveWriteThread, data, 0, NULL));
		if (thread == NULL)
		{
			logger_->error(_T("G[܂B\n"));
			logger_->debug(_T("oXbh̍쐬Ɏs܂B\n"));
		}
		else
		{
			DWORD cursor = 0;

			// oXbh̃G[m܂Ń[v
			while (!data->error)
			{
				// ǂݍ݊𔻒
				if (cursor >= streamInfo.dwLength)
				{
					complete = true;
					break;
				}

				// _̓ǂݏJ[\ʒu擾
				size_t readCursor = cursor * streamInfo.dwSampleSize;
				size_t writeCursor = data->writer->getWritten();

				// ǂݏo]n̂
				if (readCursor - writeCursor < data->bufferLength)
				{
					// ǂݏof[^obt@ɏލۂ͈̔͂擾
					size_t start = readCursor % data->bufferLength;
					size_t end = writeCursor % data->bufferLength;
					if (end <= start)
						end = data->bufferLength;

					// xɓǂݍރTCY̓obt@1/4Ɍ
					if (end - start > (data->bufferLength >> 2))
						end = start + (data->bufferLength >> 2);

					// ǂݏoTv擾
					DWORD readSamples = static_cast<DWORD>((end - start) / streamInfo.dwSampleSize);
					if (readSamples > streamInfo.dwLength - cursor)
						readSamples = streamInfo.dwLength - cursor;

					// ǂݍݏ
					LONG writeBytes;
					LONG writeSamples;
					result = AVIStreamRead(stream, cursor, static_cast<LONG>(readSamples), &data->buffer[start], static_cast<LONG>(end - start), &writeBytes, &writeSamples);
					if (result == AVIERR_BUFFERTOOSMALL)
					{
						data->error = true;
						logger_->error(_T("I[fBIXg[ǂݍ݂Ɏs܂Bǂݍ݃obt@̃TCY܂B\n"));
						break;
					}
					else if (result == AVIERR_MEMORY)
					{
						data->error = true;
						logger_->error(_T("I[fBIXg[ǂݍ݂Ɏs܂BȂȂ܂B\n"));
						break;
					}
					else if (result == AVIERR_FILEREAD)
					{
						data->error = true;
						logger_->error(_T("I[fBIXg[ǂݍ݂Ɏs܂BfBXNG[܂B\n"));
						break;
					}
					else if (FAILED(result))
					{
						data->error = true;
						logger_->error(_T("I[fBIXg[ǂݍ݂Ɏs܂BsȃG[܂B\n"));
						break;
					}

					logger_->debug(_T("%d - %d ̃f[^ǂݎ܂B\n"), cursor * streamInfo.dwSampleSize, (cursor + writeSamples) * streamInfo.dwSampleSize - 1);

					// vƈႤʂǂݍ܂ꂽG[
					// G[ƂfBXNG[͏ŃJo[Ă̂ōȂR͖͂
					if (readSamples != writeSamples || streamInfo.dwSampleSize * writeSamples != writeBytes)
					{
						data->error = true;
						logger_->error(_T("I[fBIXg[ǂݍ݂Ɏs܂Bf[^TCYɕs܂B\n"));
						break;
					}

					// ǂݍ݃J[\XV
					cursor += writeSamples;

					// oXbhQƂǂݍ݃J[\XV
					data->readCursor = cursor * streamInfo.dwSampleSize;

					// oXbh~܂Ă΋N
					SetEvent(data->writeEvent);
				}
				else
				{
					// oXbhNē҂
					SetEvent(data->writeEvent);
					WaitForSingleObject(data->readEvent, 1000);
				}
			}

			while (!data->writer->isComplete() && !data->error)
			{
				SetEvent(data->writeEvent);
				WaitForSingleObject(data->readEvent, 1000);
			}

			CloseHandle(thread);
		}
	}

	delete data;

	return complete;
}

WaveExtractor::SharedData::SharedData(WAVEFORMATEX format, size_t sampleSize, size_t length, const Logger* log)
: error(false), bufferLength(sampleSize * 8 * 1024 * 1024), readCursor(0), readEvent(NULL), writeEvent(NULL), logger(log)
{
	writer = new WaveFileWriter(format, length, log);
	buffer = new BYTE[bufferLength];
}

WaveExtractor::SharedData::~SharedData()
{
	if (writeEvent != NULL)
		CloseHandle(writeEvent);

	if (readEvent != NULL)
		CloseHandle(readEvent);

	delete [] buffer;

	writer->close();
	delete writer;
}

bool WaveExtractor::SharedData::initialize(const TCHAR* outputFilePath)
{
	readEvent = CreateEvent(NULL, FALSE, TRUE, NULL);
	if (readEvent == NULL)
	{
		logger->error(_T("G[܂B\n"));
		logger->debug(_T("ǂݍ݃XbhpCxg̍쐬Ɏs܂B\n"));
		return false;
	}

	writeEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
	if (writeEvent == NULL)
	{
		logger->error(_T("G[܂B\n"));
		logger->debug(_T("oXbhpCxg̍쐬Ɏs܂B\n"));
		return false;
	}

	return writer->open(outputFilePath);
}